"""# Standardize test format

Tool for standardizing Partisia's Java tests.

Is capable of migrating from Javadoc based property specification to the
`@DisplayName` specification and back.
"""

import re

PATTERN_JAVADOC = r'/\*\*(?P<javadoc>.*(\n\s*\*.*)*?)\s*\*/'
PATTERN_DISPLAY_NAME_EXPRESSION = r'\s*(?:"[^"]*?")(?:\s*\+\s*"[^"]*?")*\s*'
PATTERN_DISPLAY_NAME = (
    r'@DisplayName\((?P<display>' + PATTERN_DISPLAY_NAME_EXPRESSION + r')\)'
)

TEST_PATTERN: re.Pattern = re.compile(
    r''
    + r'(?:'
    + PATTERN_JAVADOC
    + r'\s*)?'
    + r'@(?P<annotation>Test|BeforeEach|ParameterizedTest)\s*'
    + r'(?:'
    + PATTERN_DISPLAY_NAME
    + r'\s*)?'
    + r'(?P<visibility>public\s+|private\s+)?void\s+(?P<name>\w+)\(',
    re.IGNORECASE,
)

END_SYMBOLS = '.?!'

IGNORABLE_WORDS_IN_NAME = frozenset(
    {
        '',
        'a',
        'the',
        'is',
        'are',
    },
)


def format_test_prefix(
    name: str,
    description: str,
    annotation: str,
    visibility: str,
    with_javadoc: bool,
    with_display_name: bool,
) -> str:
    indent = ''
    str_builder = []
    if with_javadoc:
        str_builder += ['/** ', description, ' */', '\n']
    str_builder += [indent, '@', annotation]
    if with_display_name:
        str_builder += ['\n', indent, '@DisplayName("']
        str_builder.append(description.rstrip(END_SYMBOLS))
        str_builder += ['")']
    str_builder += ['\n', indent]
    if visibility:
        str_builder += [visibility, ' ']
    str_builder += ['void ', name, '(']
    return ''.join(str_builder)


def to_camel_case_word(word: str, first: bool = False) -> str | None:
    if word.lower() in IGNORABLE_WORDS_IN_NAME:
        return None
    if len(word) > 1 and word.upper() == word:
        word = word.lower()
    return (word[0].lower() if first else word[0].upper()) + word[1:]


def to_camel_case(description: str) -> str:
    description = (
        description.replace(',', ' ')
        .replace('.', ' ')
        .replace('"', ' ')
        .replace('+', ' ')
        .strip(' \t.')
    )
    words = description.split(' ')

    words[0] = to_camel_case_word(words[0], first=True)
    for i in range(1, len(words)):
        word = to_camel_case_word(words[i])
        if not word:
            words[i] = ''
        else:
            words[i] = word

    return ''.join(words)


def from_camel_case(name: str) -> str:
    return re.sub(
        r'(^|[A-Z])[a-z]*', lambda x: ' ' + x.group(0).capitalize(), name,
    ).strip()


def parse_javadoc_to_description(text: str | None):
    if text is None:
        return ''
    text = re.sub(r'\n\s*\*', ' ', text).strip()
    text = re.sub(r'\{@link\s+(\w+)\}', r'\1', text)
    return text


def parse_display_name_to_description(text: str | None):
    if text is None:
        return ''

    text = text.strip()
    text = re.sub(r'"\s*\+\s*"', '', text)
    text = text.strip('"')
    text = text.strip()
    return text


def replace_test_pattern(
    match: re.Match, with_javadoc: bool, with_display_name: bool,
) -> str:
    javadoc = parse_javadoc_to_description(match.group('javadoc'))

    annotation = match.group('annotation').strip()
    visibility = (match.group('visibility') or '').strip()
    name = match.group('name').strip()
    display = parse_display_name_to_description(match.group('display'))

    description = display or javadoc or from_camel_case(name)

    return format_test_prefix(
        to_camel_case(description),
        description,
        annotation,
        visibility,
        with_javadoc=with_javadoc,
        with_display_name=with_display_name,
    )


IMPORT_DISPLAY_NAME = 'import org.junit.jupiter.api.DisplayName;'


def standardize_java_text(text: str, with_javadoc: bool, with_display_name: bool):
    text = TEST_PATTERN.sub(
        lambda m: replace_test_pattern(m, with_javadoc, with_display_name), text,
    )
    if '@DisplayName' in text and IMPORT_DISPLAY_NAME not in text:
        lines = text.split('\n')
        lines.insert(1, IMPORT_DISPLAY_NAME)
        text = '\n'.join(lines)
        del lines
    return text
